import numpy as np
from scipy.special import loggamma

from helperRoutines import (vech,
                            unvech,
                            fLogit,
                            fLogistic)


###############################################################################
######################    Simulation Functions        #########################
###############################################################################

###############################################################################
# setInitialMeans()
def setInitialMeans(
        iSimulationType,
        iTimeAhead,
        iT,
        mCircleCenters,
        dCircleRadius,
        vTime,
        dCircleMidpointDistance,
        iTVprob,
        dSemiMajorAxis,
        iEllipse1Orientation,
        iEllipse2Orientation):
    """
    PURPOSE
    The initial means are set based on the chosen simulation type, which can be either a
    converging circles, Y-shaped, flipped-Y or elliptical trajectories.

    INPUTS:
        iSimulationType: integer;   0 for ovals,
                                    1 for converging circles,
                                    2 for Y-shaped
                                    3 for flipped-Y.
        
        iTimeAhead: integer
            representing the time ahead for the simulation.
        
        iT: integer
            representing the total number of time steps.
        
        mCircleCenters: (2 x 2) matrix
            representing the center of the circles.
        
        dCircleRadius: float
        representing the radius of the circles.
        
        vTime: vector
            representing time steps of the simulation.
        
        dCircleMidpointDistance: float
            representing the distance between the centers of the circles.
        
        iTVprob: integer
            Representing the type of elliptical trajectory.
            if 0, transition probs are constant; 
            if 1, they are time-varying. 
            This alters the trajectories of the true means.
            
        dSemiMajorAxis: float
            representing the semi-major axis of the ellipse.
        
        iEllipse1Orientation/iEllipse2Orientation: integers
            representing the orientation of the first and second ellipse.
        
    OUTPUTS:
        mPreviousMeans: (2 x 2) matrix
            representing the initial means of the simulation.
        
        mCircleCenters: (2 x 2) matrix 
            representing the center of the circles.
    """
    if (iSimulationType == 1):
        # converging circles
        # Fix angles to half a rotation.
        vTime = 2 * np.pi * np.arange(1, iT + 1) / (2 * iT)

        mCircleCenters = np.array([[-dCircleRadius, 0.],
                                   [dCircleRadius, 0.]]
                                  )
        mPreviousMeans = mCircleCenters + dCircleRadius * np.array([[0, 1],
                                                                    [0, 1]]
                                                                   )
    elif (iSimulationType == 2):
        # Y-shaped
        # Here the radius becomes the Y distance .
        # Also, dCircleMidpointDistance becomes the X distance.
        # No more circular movement, just straight lines.

        mCircleCenters = np.array([[dCircleMidpointDistance, dCircleRadius],
                                   [-dCircleMidpointDistance, dCircleRadius]]
                                  )
        mPreviousMeans = mCircleCenters

    elif (iSimulationType == 3):
        # flipped-Y
        # Same as 2 but starts together and separates
        # Start in the same position, will split after t=T/2

        mCircleCenters = np.array([[0., dCircleRadius],
                                   [0., dCircleRadius]]
                                  )
        mPreviousMeans = mCircleCenters

    elif (iTVprob == 1):
        # Elliptical trajectories
        # breakpoint()
        mPreviousMeans = mCircleCenters + dCircleRadius * \
            np.array([[-(dSemiMajorAxis**(1 - iEllipse1Orientation)) * np.cos(0),
                        (dSemiMajorAxis**(iEllipse1Orientation)) * np.sin(0)],
                      [(dSemiMajorAxis**(1 - iEllipse2Orientation)) * np.cos(vTime[iTimeAhead]),
                        (dSemiMajorAxis**(iEllipse2Orientation)) * np.sin(vTime[iTimeAhead])]]
                      )

    else:
        # Synchronized elliptical trajectories
        mPreviousMeans = mCircleCenters + dCircleRadius * \
            np.array([[-(dSemiMajorAxis**(1 - iEllipse1Orientation)) * np.cos(0),
                       (dSemiMajorAxis**(iEllipse1Orientation)) * np.sin(0)],
                      [-(dSemiMajorAxis**(1 - iEllipse2Orientation)) * np.cos(vTime[iTimeAhead]),
                       (dSemiMajorAxis**(iEllipse2Orientation)) * np.sin(vTime[iTimeAhead])]]
                     )

    return mPreviousMeans, mCircleCenters


###############################################################################
# generateTransitionProbabilities()
def generateTransitionProbabilities(
        i,
        mClusterDistances,
        mPreviousMeans,
        mChol,
        dGamma,
        dSmooth):
    if (i == 0):
        mClusterDistances = fComputeDistanceMatrix(
            mPreviousMeans, mChol, dGamma)
    else:
        mClusterDistances = (1 - dSmooth) * mClusterDistances \
            + dSmooth * fComputeDistanceMatrix(mPreviousMeans,
                                              mChol,
                                              dGamma)
    # Compute the transition probabilities from t to t+1 ;
    # based on the means from time t
    mStateTransitionProbabilities = mapClusterMeanDistancesIntoProbabilities(
        mClusterDistances, dGamma)

    return mStateTransitionProbabilities, mClusterDistances


###########################################################
# fComputeDistanceMatrix()
def fComputeDistanceMatrix(mClusterMeans, mClusterChols, dGamma):
    """
 PURPOSE:
     Compute the Mahalanobis distance matrix between all clusters.

 INPUTS:
     mClusterMeans: matrix 
         containing the mean of each cluster
     
     mClusterChols: matrix 
         containing the Cholesky decomposition of the covariance matrix of each cluster
     
     dGamma: float
         transition probability decay parameter

 OUTPUTS:
     mDistanceMatrix: matrix
         containing the Mahalanobis distance between all clusters
 """
    iP = mClusterMeans.shape[0]

    mDistanceMatrix = np.zeros((iP, iP))
    mChol = np.tril(unvech(mClusterChols.mean(axis=0)))

    mCholInv = np.linalg.inv(mChol).T

    for i in range(iP - 1):
        for j in range(i, iP):
            mDistanceMatrix[i, j] = np.sqrt(
                (((mClusterMeans[i, :] - mClusterMeans[j, :]) @ mCholInv.T)**2).sum())
            mDistanceMatrix[j, i] = mDistanceMatrix[i, j]

    return mDistanceMatrix


###########################################################
# mapClusterMeanDistancesIntoProbabilities()
def mapClusterMeanDistancesIntoProbabilities(mDistanceMatrix, dGamma):
    """
    PURPOSE:
        Map the cluster means distances into a transition probability matrix.
        The underlying model is
            Pi(l,m) = exp(-dGamma*d(l,m)) / (1 + sum_(q\not= l) exp(-dGamma*d(l,q)))
        with d(l,q) is the distance between means of cluster l and q.

    INPUTS:
        mClusterMeans: matrix(iP x iD)
            with iP cluster means in dimension iD
            
        dGamma: float
            transition probability decay parameter

    RETURNS:
        mTransitionMatrix: (iP x iP)matrix
            contains transition probabilities.
    """
    mTransitionMatrix = np.exp(-dGamma * mDistanceMatrix)
    mTransitionMatrix = mTransitionMatrix / \
        mTransitionMatrix.sum(axis=1, keepdims=True)

    return mTransitionMatrix


###########################################################
# generateNewFirmStates()
def generateNewFirmStates(vCurrentStates, mStateTransitionProbabilities):
    """
    PURPOSE:
    Generate new firm states based on current states and state transition probabilities.
    
    INPUTS:
        vCurrentStates: 
            contains current state of firms
        
        mStateTransitionProbabilities: matrix 
            contains state transition probabilities
    
    OUTPUTS:
        mAid: matrix
            contains new firm states
    """
    mAid = mStateTransitionProbabilities.cumsum(axis=1)[vCurrentStates]
    mAid = (mAid < np.random.uniform( 0, 1, len(vCurrentStates))[:, None]).sum(axis=1)
    return mAid


###############################################################################
# updateMeans()
def updateMeans(mPreviousMeans,
                iSimulationType,
                i,
                iCluster2,
                iT,
                mCircleCenters,
                dCircleRadius,
                vTime,
                dCircleMidpointDistance,
                iTVprob,
                dSemiMajorAxis,
                iEllipse1Orientation,
                iEllipse2Orientation):
    """
    PURPOSE:
        Update the cluster means based on the current time step and the simulation type.
        
    INPUTS:
        mPreviousMeans: (2 x 2) matrix
            representing the previous means before update
            
        i: integer
            current time step
        
        iCluster2: integer
            cluster index
        
        iSimulationType: integer;   0 for ovals,
                                    1 for converging circles,
                                    2 for Y-shaped
                                    3 for flipped-Y.
        
        iT: integer
            representing the total number of time steps.
        
        mCircleCenters: (2 x 2) matrix
            representing the center of the circles.
        
        dCircleRadius: float
        representing the radius of the circles.
        
        vTime: vector
            representing time steps of the simulation.
        
        dCircleMidpointDistance: float
            representing the distance between the centers of the circles.
        
        iTVprob: integer
            Representing the type of elliptical trajectory.
            if 0, transition probs are constant; 
            if 1, they are time-varying. 
            This alters the trajectories of the true means.
            
        dSemiMajorAxis: float
            representing the semi-major axis of the ellipse.
        
        iEllipse1Orientation/iEllipse2Orientation: integers
            representing the orientation of the first and second ellipse.
        
    OUTPUTS:
        mPreviousMeans: (2 x 2) matrix 
            updated cluster means
    """
    if (iSimulationType == 1):
        if(i < iT / 2):  # Else nothing, just stay put
            mPreviousMeans = mCircleCenters + dCircleRadius * \
                np.array([[np.sin(vTime[i]), np.cos(vTime[i])],
                          [-np.sin(vTime[i]), np.cos(vTime[i])]]
                         )

    elif (iSimulationType == 2):
        # Take linear steps
        dCurrentX = ((i + 1 < iT / 2) * 2 * (i + 1) * dCircleMidpointDistance /
                     iT) + ((i + 1 >= iT / 2) * dCircleMidpointDistance)

        mPreviousMeans = mCircleCenters + \
            np.array([[-dCurrentX, -(i + 1) * 2 * dCircleRadius / iT],
                      [dCurrentX, -(i + 1) * 2 * dCircleRadius / iT]]
                     )

    elif (iSimulationType == 3):
        # Take linear steps        
        dCurrentX = ((i+1 >= iT/2)
                     * 2 * (i+1 - iT/2)
                     * dCircleMidpointDistance 
                     /iT)

        mPreviousMeans = mCircleCenters + \
            np.array([[-dCurrentX, -(i + 1) * 2 * dCircleRadius / iT],
                      [dCurrentX, -(i + 1) * 2 * dCircleRadius / iT]]
                     )

    elif (iTVprob == 1):
        mPreviousMeans = mCircleCenters + dCircleRadius * \
            np.array([[-(dSemiMajorAxis**(1 - iEllipse1Orientation)) * np.cos(vTime[i]),
                       (dSemiMajorAxis**(iEllipse1Orientation)) * np.sin(vTime[i])],
                      [(dSemiMajorAxis**(1 - iEllipse2Orientation)) * np.cos(vTime[iCluster2]),
                       (dSemiMajorAxis**(iEllipse2Orientation)) * np.sin(vTime[iCluster2])]]
                     )

    else:
        mPreviousMeans = mCircleCenters + dCircleRadius * \
            np.array([[-(dSemiMajorAxis**(1 - iEllipse1Orientation)) * np.cos(vTime[i]),
                       (dSemiMajorAxis**(iEllipse1Orientation)) * np.sin(vTime[i])],
                      [-(dSemiMajorAxis**(1 - iEllipse2Orientation)) * np.cos(vTime[iCluster2]),
                       (dSemiMajorAxis**(iEllipse2Orientation)) * np.sin(vTime[iCluster2])]]
                     )
    return mPreviousMeans


###############################################################################
######################  Initialization Functions   ############################
###############################################################################

###############################################################################
# InitializeTimeVaryingParameterStructure()
def InitializeTimeVaryingParameterStructure(iT, iP, iK, dScalar=0):
    """
    PURPOSE:
        Initiate a 3d structure filled with zeros (or an optional scalar).

    INPUTS:
        iT	: size of the array
        iP	: rows of the array elements
        iK	: columns of the array elements
        dScalar: [optional] if provided, a scalar, otherwise set to 0

    OUTPUTS:
        iT x iP x iK structure filled with dScalar

    """
    try:
        mTmpParameters = np.zeros((iT, iP, iK))
        for i in range(iT):
            mTmpParameters[i, :, :] += dScalar

    except TypeError:
        print("Check if you actually put through an integer value as input")
        iT = int(iT)
        iP = int(iP)
        iK = int(iK)
        mTmpParameters = np.zeros((iT, iP, iK))
        for i in range(iT):
            mTmpParameters[i, :, :] += dScalar

    return mTmpParameters
        


###############################################################################
######################    Maximum Likelihood functions        #################
###############################################################################

###########################################################
# fRestrictPar()
def fRestrictPar(cClass, vP, iMeanSpecificSmoothParam):
    """
    PURPOSE:
        Restrict certain parameters of vP. 
        Note: currently only there for didactical purposes, no transformation
        is specified.
    
    INPUTS:
        cClass: cSimulationStudy() object
            Contains storage matrices for data, parameters and states.
            See earlier descriptions for details.
            
        vP : vector
            vector of parameters; dimensiosn may vary
        
        iMeanSpecificSmoothParam: integer
            flag for mean-specific smoothing parameter
        
    OUTPUTS:
        vPar: vector
            modified version of input parameter vector vP
    """
    idx = 0
    vPar = vP
    if(iMeanSpecificSmoothParam == 1):
        # For didactical purposes we use -1 +1
        idx = np.arange(0, cClass.s_ammData.shape[1] - 1 + 1)

    return vPar


###########################################################
# fGetParameters()
def fGetParameters(cClass, vP, iMeanSpecificSmoothParam,
                   dInverseDegreesOfFreedom):
    
    """
    Purpose:
        Extract parameters from the passed in parameter array and transform them as necessary.
    
    Inputs:
        cClass: cSimulationStudy() object
            Contains storage matrices for data, parameters and states.
            See earlier descriptions for details.
            
        vP : vector
            vector of parameters; dimensiosn may vary
        
        iMeanSpecificSmoothParam: integer
            flag for mean-specific smoothing parameter
            
        dInverseDegreesOfFreedom: float
            dof in case of Student-t mixtures
    
    Outputs:
        vEwma, (iD+1 x 1) vector
            parameter with iD smoothing parameters for the means
            and 1 smoothing parameter for variances
        
        dGamma, float
            Decay parameter for transition probabilities
        
        dDegreesOfFreedom/dInverseDegreesOfFreedom: float
            dof in case of Student-t mixtures
    
    """
    ix = 0

    # transform the parameters
    vP = fRestrictPar(cClass,
                      vP,
                      iMeanSpecificSmoothParam)

    # get the dimensions
    iD =  2 #cClass.s_ammData.shape[1]

    # first element(s) vP: smoothing parameter(s) for the mean
    if(iMeanSpecificSmoothParam == 1):
        vEwma = vP[ix:ix + iD - 1 + 1]
        ix += iD
    else:
        vEwma = np.ones(iD) * vP[ix:ix + 1]
        ix += 1

    # no element for: smoothing parameter(s) for the covariance matrix
    vEwma = np.append(vEwma, 0)

    # next element vP: degrees of freedom
    if (np.isnan(dInverseDegreesOfFreedom) or dInverseDegreesOfFreedom == 0):
        dInverseDegreesOfFreedom = np.nan
        dDegreesOfFreedom = np.nan

    elif (dInverseDegreesOfFreedom > 0):
        dInverseDegreesOfFreedom = 0.5 * fLogistic(vP[ix])
        ix += 1
        dDegreesOfFreedom = 1.0 / dInverseDegreesOfFreedom

    else:
        dDegreesOfFreedom = 1.0 / np.abs(dInverseDegreesOfFreedom)
        ix += 1

    dGamma = np.exp(vP[ix])
    ix += 1

    return vEwma, dGamma, dDegreesOfFreedom, dInverseDegreesOfFreedom 


###########################################################
# fKMeansClustering()
def fKMeansClustering(mData, iNumberOfClusters, iNumberOfTries=10):
    """
    PURPOSE:
        Perform K-means clustering.
        
    INPUTS:
    mData: (iN x iD) 2D matrix
        containing the data to be clustered
        
    iNumberOfClusters: integer
        number of clusters
        
    iNumberOfTries: integer
        the number of times the algorithm is run from different random 
        initial values/cluster assignments

    OUTPUTS:
    mParameters: iNumberOfClusters x iD(ID+3)/2 matrix, rows for each cluster,
                each row first containing the means, and then the vech of the
                  choleski of the covariance matrix
                  
    g(mAssignments): (iN x iNumberOfClusters) matrix
        Containing in (i,j) a
            1, if unit i belongs to cluster j
            0, else
    """
    iN = mData.shape[0]
    iD = mData.shape[1]
    dMinimumMSE = np.nan
    mAssignments = np.zeros(iN)

    for i in range(iNumberOfTries):
        # initialize menas and variances
        mAssignmentsTmp = np.floor(np.random.uniform(
            0, 1, iN) * iNumberOfClusters).astype(int)
        mAssignmentsPrevious = np.zeros(len(mAssignmentsTmp))
        iClustersAreChanging = 1

        # iterate until cluster assignment is stable/has converged
        while(iClustersAreChanging != 0):
            # compute the mean
            mAid = np.identity(iNumberOfClusters)[mAssignmentsTmp, :]
            mMeans = mAid.T @ mData / mAid.sum(axis=0)[:, None]

            # compute MSE
            dMinimumMSETmp = ((mData - mMeans[mAssignmentsTmp, :])**2).sum()

            # update cluster assignments
            mAid = np.zeros((iN, iNumberOfClusters))
            for j in range(iNumberOfClusters):
                mAid[:, j] = ((mData - mMeans[j, :])**2).sum(axis=1)

            mAssignmentsTmp = np.argmin(mAid, axis=1)

            iClustersAreChanging = np.abs(
                mAssignmentsTmp - mAssignmentsPrevious).sum()
            mAssignmentsPrevious = mAssignmentsTmp

        # update to new minimum if the fit is better
        if (np.isnan(dMinimumMSE) or dMinimumMSETmp < dMinimumMSE):
            dMinimumMSE = dMinimumMSETmp
            mParameters = mMeans
            mAssignments = mAssignmentsTmp

    # compute covariances
    mParameters = np.column_stack((mParameters, np.zeros(
        (iNumberOfClusters, int((iD * (iD + 1)) / 2)))))
    
    for i in range(iNumberOfClusters):
        mParameters[i, iD:] = vech(np.linalg.cholesky(
            np.cov(mData[mAssignments == i].T)))
    

    return mParameters, np.identity(iNumberOfClusters)[mAssignments, :]


###########################################################
# fInitializeVp()
def fInitializeVp(cClass, iN, iP, iD, dInverseDegreesOfFreedom,
                  iMeanSpecificSmoothParam, dWarmStart):
    """
    PURPOSE:
        Initialize the smoothing parameter vector and cluster assignments.

    INPUTS:
        cClass: cSimulationStudy() object
            Contains storage matrices for data, parameters and states.
            See earlier descriptions for details.
            
        iN: integer
            number of firms per mixture component
            
        iP: integer
            number of mixture components
            
        iD: integer
            number of characteristics
            
        dInverseDegreesOfFreedom: float
            dof in case of Student-t mixtures
        
        iMeanSpecificSmoothParam: integer
            flag for mean-specific smoothing parameter
        
        dWarmStart: float
            starting value for the smoothing parameter
        
    OUTPUTS:
        vP: 1D vector
            smoothing parameter vector
            
        mClusterParameters: 2D vector
            cluster parameters bt KKN
    """
    # mean smoothing paramter
    if (iMeanSpecificSmoothParam == 1):
        vP = np.ones(iD) * dWarmStart
    else:
        vP = np.array([dWarmStart])

    # starting predicted cluster probabilities as K-Means Clustering
    # assignments
    mClusterParameters, cClass.s_ammPredictedClusterProbabilities[0,:,:] = fKMeansClustering(
        cClass.s_ammData[:, :, 0], iP)
    
    cClass.s_mInitialPredictedProbabilities = cClass.s_ammPredictedClusterProbabilities[0,:,:]

    # degrees of freedom
    if (dInverseDegreesOfFreedom > 0):
        mAid = fLogit(2 * dInverseDegreesOfFreedom)
        vP = np.hstack((vP, mAid))

    # transition probability decay parameter
    vP = np.hstack((vP, np.log(3)))

    return vP, mClusterParameters


###########################################################
# fGetMeanAndCovarianceMatrixFromParameter()
def fGetMeanAndCovarianceMatrixFromParameter(vParameters, iD, dRegularizeCovariance=0):
    """
    PURPOSE:
        Extract mean and covariance matrices from a single given vector.
    
    INPUTS:
        vParameters: 1D vector
            vector of parameters
            
        iD: integer
            number of dimensions
            
        dRegularizeCovariance: [optional] scalar
            regularization term for the covariance matrix. 
            Default is 0.
    
    OUTPUTS:
        vMean: 1D vector
            vector of means
            
        mCov: 2D vector
            covariance matrix
            
        mChol: 2D vector
            Cholesky decomposition of the covariance matrix
    """
    ix = 0
    vMean = vParameters[int(ix):int(ix+iD-1+1)]
    ix += iD
    mChol = np.tril(unvech(vParameters[int(ix):int(ix+iD*(iD+1)/2-1)+1])) 
    mCov = mChol @ mChol.T + dRegularizeCovariance*np.identity(iD)
    ix += iD*(iD+1)/2
    mChol = np.linalg.cholesky(mCov)
    
    return vMean, mCov, mChol
    

###########################################################
# fHMMGasFilter()  
def fHMMGasFilter(cClass, vPar, dGamma, mInitialPredictedClusterProbabilities,
                  dSmooth, dRegularizeCovariance, dInverseDegreesOfFreedom):
    
    """
    PURPOSE:
        Run the GAS filter for the mean and the covariance matrix.
        At the moment, the filter is a single parameter EWMA filter
    
    INPUTS:
        cClass: cSimulationStudy() object
            Contains storage matrices for data, parameters and states.
            See earlier descriptions for details.
            
        vPar	: (iD+1)x1 vector
            The GAS smoothing parameter for mean (0..iD-1) and variance (iD element).
            
        dGamma: float
            The rate of decay of the transition probabilities from one cluster 
            to the other in terms of the distance between cluster means, using
            Pi(l,m) = exp(-dGamma*d(l,m)) / (1 + sum_(q\not= l) exp(-dGamma*d(l,q)))
            and d(l,q) is the distance between means of cluster l and q.
            
        mInitialPredictedClusterProbabilities: (iT, iD*iN, iNrMixtureComponentsForEstimation) matrix
            Intialized cluster probabilities
            
        dSmooth: float    
            Smoothing fraction used for estimation.
                 
        dRegularizeCovariance: float    
            This number times a unit matrix is added to the covariance matrices
            to prevent degenerate results
            
        dInverseDegreesOfFreedom: float
            dof in case of Student-t mixtures
    
    OUTPUTS:
        vLik: (iT x 1) vector
            Contains the likelihood contributions at each time point t=1,...,iT       
    """
    # initialize
    iN = cClass.s_ammData.shape[0]        # number of time points
    iD = cClass.s_ammData.shape[1]        # number of cross sectional units
    iT = cClass.s_ammData.shape[2]        # number of characteristics in data
    iP = cClass.s_mParameters.shape[0]    # number of mixture components
   
    
    mLikTimeT = np.zeros((iN, iP))  # matrix of likelihood contributions at time t firm i cluster j
    vLik = np.zeros(iT)             # vector of likelihood contributions for each time t 
         
    mPredictedClusterProbabilities = mInitialPredictedClusterProbabilities
    mParameters = cClass.s_mParameters.copy()
    
    # run the filter per time step
    for i in range(iT):
        # store the time varying mean and time varying scale parameters
        cClass.s_ammParameters[i,:,:] = mParameters.copy()
        cClass.s_ammPredictedClusterProbabilities[i,:,:] = mPredictedClusterProbabilities.copy()
        
        for j in range(iP):
            # get out the parameters
            vMean, mCov, mChol = fGetMeanAndCovarianceMatrixFromParameter(mParameters[j,:].copy(), iD, dRegularizeCovariance)
            mCovInv = np.linalg.inv(mCov)
            
            if np.isnan(vMean).any():
                return np.array([-999999]*iT)
          
            # using the filtered cluster probabilities, compute the log of the cluster j likelihood contribution
            if (np.isnan(dInverseDegreesOfFreedom) or dInverseDegreesOfFreedom == 0):
                # Gaussian case
                mLikTimeT[:,j] = (-0.5*iD*np.log(2*np.pi)
                                  - np.log(np.diag(mChol)).sum()
                                  - (((cClass.s_ammData[:,:,i]-vMean) @ mCovInv) * 
                                    (cClass.s_ammData[:,:,i] - vMean)).sum(axis=1)
                                  )
            else:
                # Student-t case
                mLikTimeT[:,j] = (loggamma(0.5*(1/dInverseDegreesOfFreedom + iD))
                                  - loggamma(0.5/dInverseDegreesOfFreedom)
                                  - 0.5*iD*np.log(2*np.pi)
                                  - np.log(np.diag(mChol)).sum()
                                  - 0.5*(1/dInverseDegreesOfFreedom + iD) *
                                  np.log(1 + np.abs(1/dInverseDegreesOfFreedom) *
                                      ((cClass.s_ammData[:,:,i]-vMean) @ mCovInv * (cClass.s_ammData[:,:,i]-vMean)).sum(axis=1))
                                  )
                
        # Compute log likelihood contributions across states for time t.
        # This is the log of the weighted sum of the exp(log likelihoods for time t and cluster j).    
        # The addition is numerically sensitive, and  therefore we use a standard
        # numerical scaling 'trick'.
        vAid = np.max(mLikTimeT, axis=1)
   
        # intermediate storage object; to be updated later
        # for numerical reasons, subtract the largest component log likelihood for   
        # each individual from the log likelihood values before exponentiation
        mFilteredClusterProbabilities = np.exp(mLikTimeT - vAid[:,None]) * mPredictedClusterProbabilities
        # mFilteredClusterProbabilities= mFilteredClusterProbabilities/mFilteredClusterProbabilities.sum(axis=1)[:,None]
        
        # sum across individuals
        # for numerical reasons, subtract the largest component likelihood for
        # each individual from the log likelihood values before exponentiation
        vLik[i] = (vAid + np.log(mFilteredClusterProbabilities.sum(axis=1))).sum()
        
        # compute the filtered probabilities, to be used in the GAS update later on
        mFilteredClusterProbabilities = (mFilteredClusterProbabilities / 
                                         mFilteredClusterProbabilities.sum(axis=1)[:,None])
        
        
        cClass.s_ammFilteredClusterProbabilities[i,:,:] = mFilteredClusterProbabilities.copy()

        # Compute cluster distances        
        if (i == 0):
            mClusterDistances = fComputeDistanceMatrix(mParameters[:,:iD].copy(),
                                                       mParameters[:,iD:].copy(),
                                                       dGamma)
        else:
            mClusterDistances = (1-dSmooth)*mClusterDistances + \
                dSmooth * fComputeDistanceMatrix(mParameters[:,:iD].copy(),
                                                 mParameters[:,iD:].copy(),
                                                 dGamma)
                
        # compute the transition probabilities from t to t+1 based on the
        # means from time t
        mPi = mapClusterMeanDistancesIntoProbabilities(mClusterDistances, dGamma)
        

        
        # loop over clusters to update the cluster means and variances
        for j in range(iP):
            # get out the parameters
            vMean, mCov, mChol = fGetMeanAndCovarianceMatrixFromParameter(mParameters[j,:].copy(), #[::-1]
                                                                          iD,
                                                                          dRegularizeCovariance
                                                                          )
            mCovInv = np.linalg.inv(mCov)

            # compute the mean's score step and update the mean
            vScore = (mFilteredClusterProbabilities[:,j].T @ (cClass.s_ammData[:,:,i] - vMean) 
                      / mFilteredClusterProbabilities[:,j].sum()
                      )
            vMean += vPar[:iD] * vScore

            # compute the covariance matrix's score step and update the covariance matrix
            mScore = ((cClass.s_ammData[:,:,i] - vMean).T
                      @ (mFilteredClusterProbabilities[:,j,None] * (cClass.s_ammData[:,:,i] - vMean))
                      / mFilteredClusterProbabilities[:,j].sum()
                      - mCov
                      )
            
            mCov += vPar[iD] * mScore  # ask why vPAr[iD] is always 0
            if np.isnan(mParameters).any():
                return np.array([-999999]*iT)

            # store parameters into matrix object
            ix = 0
            mParameters[j, int(ix):int(ix+iD-1+1)] = vMean.copy()

            ix += iD
            mParameters[j, int(ix):int(ix+iD*(iD+1)/2)] = vech(mChol).copy()
            ix += iD*(iD+1)/2
            
        # update predicted probabilities for next period
        mPredictedClusterProbabilities = mFilteredClusterProbabilities @ mPi

    return vLik