###########################################################
# Imports
import numpy as np


###############################################################################
##########################    Helper Routines      ############################
###############################################################################


###############################################################################
### fVecr3dStructure()
def fVecr3dStructure(mX):
    """
    PURPOSE: 
    Put the 3d structure mX into vec() form over the rows.
    
    INPUTS:
    mX: iT x iP x iK array.
    
    OUTPUTS:
    vP: 	1 x (iT * iP * iK) vector 
    """
    # initialize
    vAid = mX[0,:,:].reshape(1,-1)
    vP = vAid.reshape(vAid.shape[1])
    for i in range(1, mX.shape[0]):
        vAid = mX[i,:,:].reshape(1,-1)
        vP=np.hstack((vP, vAid.reshape(vAid.shape[1],)))
        
    return vP


###############################################################################
### fLogit()
def fLogit(mX):
    # Undoes logistic transformation (0,1) -> R
    try:
        return np.log(mX) - np.log(np.ones(mX.shape) - mX)
    except AttributeError:
        return np.log(mX) - np.log(1 - mX)


###############################################################################
### fLogistic()
def fLogistic(vX):
    # Logistic transformation R -> (0,1)
    try:
        return np.choose(vX > 0,
                         [np.ones(vX.shape) / (np.ones(vX.shape) + np.exp(-vX)),
                          np.exp(vX) / (np.ones(vX.shape) + np.exp(vX))])
    except AttributeError:
        return np.choose(vX > 0,
                         [1 / (1 + np.exp(-vX)),
                          np.exp(vX) / (1 + np.exp(vX))]
                         )
    
    
###############################################################################
### vech()
def vech(mMatrix):
    """
    PURPOSE:
        Create vector by stacking lower diagional entries(columns)
        per the vech() operator.

    INPUTS:
        mMatrix, matrix(m x n), matric we want to perform the VECH() operation on.

    RETURNS:
        VECH(mMatrix), (m(m+1)/2 - max(m-n,0)(max(m-n,0)+1)/2) vector

    """

    return mMatrix.T[np.triu_indices_from(mMatrix)]


###############################################################################
### unvech()
def unvech(vX):
    """
    PURPOSE:
        Create vector by stacking lower diagional entries(columns)
        per the vech() operator.
        ! Taken directly from the Statsmodels source code 
        ! Only put in Hungarian notation.

    INPUTS:
        vX,     vector of size n*(n-1)/2, represent lower triangular elements

    RETURNS:
        mResult, matrix(nxn) with vX on the lower triangular elements.

    """
    # quadratic formula, correct fp error
    iRows = 0.5 * (-1 + np.sqrt(1 + 8 * len(vX)))
    iRows = int(np.round(iRows))

    mResult = np.zeros((iRows, iRows))
    mResult[np.triu_indices(iRows)] = vX
    mResult = mResult + mResult.T

    # divide diagonal elements by 2
    mResult[np.diag_indices(iRows)] /= 2

    return mResult