# -*- coding: utf-8 -*-
###########################################################
# Imports
import numpy as np
import csv
from mainFunctions_rewritten import (fInitializeImportantMatrices,
                                     fSimulateData,
                                     fMaximumLikelihood)

import concurrent.futures


###########################################################
# Classes
class cSimulationStudy:
    def __init__(self):
        self.s_ammData = []
        self.s_ammParameters = []
        self.s_ammTrueParameters = []
        self.s_mTrueStates = []
        self.s_ammPredictedClusterProbabilities = []
        self.s_ammFilteredClusterProbabilities = []
        self.s_mInitialPredictedProbabilities = []
        self.s_mParameters = []
        


def parrallelSimulation(iSim, cSimul, dictSimulationParams, dictEstimationParams):
    # sign of life
    if(iSim % 1 == 0):
        print('\n============================')
        print(f'Curent simulation : {iSim}')
        print('============================\n')

    # Simulation 
    fSimulateData(cSimul, dictSimulationParams)

    # Maximum Likelihood Estimation
    vResult = fMaximumLikelihood(cSimul, dictEstimationParams)

    return vResult
        

###########################################################
# main()
def main():
    #########################################
    # Magic Numbers
    #########################################
    iRanSeed = 1234
    iNrSimulations =250
    sResultsPath = "results"
    s_iPrint = 0
    iParrallel = False
    iSimulationType = 3             # 0 for ovals
                                    # 1 for converging circles
                                    # 2 for Y-shaped
                                    # 3 for flipped-Y.

    # cluster arguments
    dInverseDegreesOfFreedomForSimulation = 0
    dInverseDegreesOfFreedomForEstimation = 0
    iNrMixtureComponentsForSimulation = 2
    iNrMixtureComponentsForEstimation = 2
    dCircleRadius = 2
    dCircleMidpointDistance = 2
    iNrLaps = 1
    
    # cluster dimension arguments
    iT = 40
    iD = 2
    iN = 100        # iN is number of firms PER MIXTURE COMPONENT;
                    # total nr. firms is iN*iNrMixtureComponentsForSimulation


    dGamma = 0.25
    iTVprob = 1	            # if 0, transition probs are constant;
                            # if 1, they are time-varying.
                            # This alters the trajectories of the true means.
    # geometric arguments
    dSimSemiMajorAxis = 2
    dSimVarianceComponent1 = 1
    dSimVarianceComponent2 = 1
    dCovariance = 0.9
    iEllipse1Orientation = 1        # 1 for standing;
    iEllipse2Orientation = 0        # 0 for lying down
    dSmoothEstimation = 0.25
    dSmoothDGP = 0.1

    # estimation related arguments
    iNrStartingValues = 1
    dWarmStart = 1.7
    dRegularizeCovariance = 0
    iMeanSpecificSmoothParam = 0	 # if 1, a smoothing parameter per mean
                                    # else single smoothing parameter for all
    iBurn = 0
    
    #########################################
    # Initialization
    #########################################
    mResult = None
    mCorr = np.array([[dSimVarianceComponent1, -dCovariance],
                      [-dCovariance, dSimVarianceComponent2]])

    # Create Paramter dictionaries
    dictSimulationParams = {0: iN,
                            1: iD,
                            2: iNrMixtureComponentsForSimulation,
                            3: dInverseDegreesOfFreedomForSimulation,
                            4: iT,
                            5: mCorr,
                            6: dCircleMidpointDistance,
                            7: dCircleRadius,
                            8: iNrLaps,
                            9: dSmoothDGP,
                            10: iTVprob,
                            11: dGamma,
                            12: dSimSemiMajorAxis,
                            13: iEllipse1Orientation,
                            14: iEllipse2Orientation,
                            15: dSmoothDGP,
                            16: iSimulationType,
                            }

    dictEstimationParams = {0: iNrMixtureComponentsForEstimation,
                            1: iNrStartingValues,
                            2: s_iPrint,
                            3: dInverseDegreesOfFreedomForEstimation,
                            4: dSmoothEstimation,
                            5: iMeanSpecificSmoothParam,
                            6: dWarmStart,
                            7: dRegularizeCovariance,
                            8: iBurn
                            }
    
    

    # Initialize major data and paramter matrices
    cSimul = cSimulationStudy()
    
    (cSimul.s_ammParameters,
     cSimul.s_ammTrueParameters,
     cSimul.s_mTrueStates,
     cSimul.s_ammPredictedClusterProbabilities,
     cSimul.s_ammFilteredClusterProbabilities,
     cSimul.s_ammData) = fInitializeImportantMatrices(iT,
                                                        iN,
                                                        iD,
                                                        iNrMixtureComponentsForEstimation,
                                                        iNrMixtureComponentsForSimulation)
                                                      
 
                                                      
                                                    

    #########################################
    # Simulation and Estimation
    #########################################
    # First check whether user has defined a seed to be used.
    try:
        np.random.seed(iRanSeed)
        print(f'We are setting the random seed to {iRanSeed}.')
    except NameError:
        print("Guess we don't want to use a seed?")
        
    
    if iParrallel:
        with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
            # Create a list of tasks to be executed concurrently
            tasks = [executor.submit(parrallelSimulation, iSim, cSimul, dictSimulationParams, dictEstimationParams)
                      for iSim in range(iNrSimulations)]
    
        # Wait for all tasks to complete and collect the results
        mResult = [task.result() for task in tasks] 
        mResult = np.vstack(mResult)
    else:     
        for iSim in range(iNrSimulations):

            # sign of life
            if(iSim % 1 == 0):
                print('\n============================')
                print(f'Curent simulation : {iSim}')
                print('============================\n')
                
            #########################################
            # Simulation
            #########################################
            fSimulateData(cSimul, dictSimulationParams)
                           
            #########################################
            # Maximum Likelihood Estimation
            #########################################
            vResult= fMaximumLikelihood(cSimul, dictEstimationParams)    
            if (iSim == 0):
                mResult = [vResult]
            else:
                mResult = mResult + [vResult] 
        mResult = np.vstack(mResult).T
 
        
        
    
    # Set the file name
    sName = f"Sim_{sResultsPath}_{iSimulationType}_{iN}_{iT}_{dGamma}_\
{dCircleRadius}_{dCircleMidpointDistance}_\
{dInverseDegreesOfFreedomForSimulation}_\
{dInverseDegreesOfFreedomForEstimation}_\
{iNrMixtureComponentsForEstimation}_\
{iNrMixtureComponentsForSimulation}_\
{iTVprob}_{iMeanSpecificSmoothParam}_{dSimSemiMajorAxis}_\
{iEllipse1Orientation}_{iEllipse2Orientation}\
_{dSimVarianceComponent1}_{dSimVarianceComponent2}_{dCovariance}_\
{dSmoothEstimation}_{dSmoothDGP}.csv"
    
    # Open the file in write mode
    with open(sName, "w", newline="") as file:
        # Create a CSV writer
        writer = csv.writer(file)
    
        # Write the data to the file
        writer.writerows(mResult)
    
    print("\n============================")
    print("END OF PROGRAM")
    print("============================\n")
    print(f"\n Results matrix mresult is stored under the name :\n {sName}")
    

if __name__ == "__main__":
    main()
    