###########################################################
# Imports
import numpy as np
from scipy.optimize import minimize

from auxiliaryFunctions_rewritten import (setInitialMeans,
                                          generateTransitionProbabilities,
                                          generateNewFirmStates,
                                          updateMeans,
                                          InitializeTimeVaryingParameterStructure,
                                          fRestrictPar,
                                          fGetParameters,
                                          fInitializeVp,
                                          fHMMGasFilter)

from helperRoutines import (vech, fVecr3dStructure)

###############################################################################
######################    Initialization Functions    #########################
###############################################################################

###############################################################################
### fInitializeImportantMatrices()
def fInitializeImportantMatrices(iT, iN, iD, iNrMixtureComponentsForEstimation, 
                                iNrMixtureComponentsForSimulation):
    
    """
    PURPOSE:
        Initialize important matrices for the simulation and estimation process.
    
    INPUTS:
        iT: integer
            number of time periods
        
        iN: integer
            number of firms per mixture component
        
        iD: integer
            number of firm characteristics
        
        iNrMixtureComponentsForEstimation: integer
            number of mixture components for estimation
        
        iNrMixtureComponentsForSimulation: integer
            number of mixture components for simulation
        
    OUTPUTS:
        s_ammParameters: 3D matrix
            parameters for estimation
        
        s_ammTrueParameters: 3D matrix
            true parameters for simulation
        
        s_mTrueStates: 2D matrix
            true states for simulation
        
        s_ammPredictedClusterProbabilities: 3D matrix
            predicted cluster probabilities for estimation
        
        s_ammFilteredClusterProbabilities: 3D matrix
            filtered probabilities for estimation
                                                
        s_ammData: 3D matrix
            data for simulation
    """
    s_ammParameters = InitializeTimeVaryingParameterStructure(iT,
                                                              iNrMixtureComponentsForEstimation,
                                                              0.5*iD*(iD+3)
                                                              )
    s_ammTrueParameters = InitializeTimeVaryingParameterStructure(iT,
                                                                  iNrMixtureComponentsForSimulation,
                                                                  0.5*iD*(iD+3)
                                                                  )
    s_mTrueStates = np.zeros((iT,iNrMixtureComponentsForSimulation*iN))
    
    # InitializationiNrMixtureComponentsForEstimation
    s_ammPredictedClusterProbabilities = InitializeTimeVaryingParameterStructure(iT,
                                                                                 iD*iN,
                                                                                 iNrMixtureComponentsForEstimation,
                                                                                 1.0/iNrMixtureComponentsForEstimation
                                                                                 )
    s_ammFilteredClusterProbabilities = InitializeTimeVaryingParameterStructure(iT,
                                                                                iD*iN,
                                                                                iNrMixtureComponentsForEstimation,
                                                                                1.0/iNrMixtureComponentsForEstimation)
    
    s_ammData = np.zeros((iN*iNrMixtureComponentsForSimulation, iD, iT))
        
    return (s_ammParameters, s_ammTrueParameters, s_mTrueStates,
            s_ammPredictedClusterProbabilities, s_ammFilteredClusterProbabilities,
            s_ammData)



###############################################################################
######################    Simulation Functions        #########################
###############################################################################

###########################################################
### fSimulateData()
def fSimulateData(cClass, dictSimulationParams):
    """ 
    PURPOSE:
        Simulate data in concentric circles with increased midpoint distance.
        Update the True States, True Parameters and generated data
        inside the cClass entries.
    
    INPUTS:
        dictSimulationParams: dictionary
            containing params used for simulation.

    RETURNS:
      cClass: cSimulationStudy() object; method fills the following:
            
        s_ammTrueParameters: 3D matrix
            true parameters for simulation
        
        s_mTrueStates: 2D matrix
            true states for simulation   
                                     
        s_ammData: 3D matrix
            simulated data for estimation
    """  
    # Initilize simulation paramters     
    iN = dictSimulationParams[0]
    iD = dictSimulationParams[1] # = 2
    iP = dictSimulationParams[2]
    dInverseDegreesOfFreedom = dictSimulationParams[3]
    iT = dictSimulationParams[4] 
    mCorr = dictSimulationParams[5]
    dCircleMidpointDistance = dictSimulationParams[6]
    dCircleRadius = dictSimulationParams[7]
    iNrLaps = dictSimulationParams[8]
    dSmooth = dictSimulationParams[9]
    iTVprob = dictSimulationParams[10]
    dGamma = dictSimulationParams[11]   # = 50
    dSemiMajorAxis = dictSimulationParams[12]   # = 1
    iEllipse1Orientation = dictSimulationParams[13]   # = 1
    iEllipse2Orientation = dictSimulationParams[14]   # = 1
    dSmooth = dictSimulationParams[15] # = 0.1
    iSimulationType = dictSimulationParams[16]  # force 0 
    
    mClusterDistances = None
    iTimeAhead = int(np.ceil(iT/4))  # factor by how much the second cluster 
                                     # is ahead of the first.
                                     
    vTime = iNrLaps * 2*np.pi * np.arange(1, iT + iTimeAhead + 1) / iT
    vCurrentStates = np.kron(np.arange(0, iP-1 +1), np.ones(iN)).astype(int) 
    
    mChol = np.linalg.cholesky(mCorr).T
    
    # change sign off-diagonal elements
    mChol2 = np.transpose(np.linalg.cholesky((mCorr * (2 * np.eye(iD) - np.ones((iD, iD))))))
    
    # initialize (previous) means
    mCircleCenters = np.array([[0., 0.],
                               [dCircleMidpointDistance, 0.]]
                              ) 

    # Depending on what type of simulation is desired, generate means
    mPreviousMeans, mCircleCenters = setInitialMeans(iSimulationType,
                                     iTimeAhead,
                                     iT,
                                     mCircleCenters,
                                     dCircleRadius,
                                     vTime,
                                     dCircleMidpointDistance,
                                     iTVprob,
                                     dSemiMajorAxis,
                                     iEllipse1Orientation,
                                     iEllipse2Orientation
                                     )

    for i in range(iT):
        # To change the starting position of cluster 2, pretend that it's in the future
        iCluster2 = i + iTimeAhead
        cClass.s_ammTrueParameters[i,:,:] = np.column_stack((mPreviousMeans, 
                                                      np.row_stack((vech(mChol).T, vech(mChol2).T))
                                                      ))      
        # Store true paramters
        cClass.s_mTrueStates[i,:] = vCurrentStates
               
        # First, generate the firm observations
        mAid = (vCurrentStates[:,None] * np.random.randn(iN*iP,iD) @ mChol) + \
            ((1-vCurrentStates)[:,None] * np.random.randn(iN*iP,iD) @ mChol2)
        
        # store the data for this cluster
        cClass.s_ammData[:, :, i] = mPreviousMeans[vCurrentStates, :] + mAid
        
        # Second, generate transition probability matrix
        mStateTransitionProbabilities, mClusterDistances = \
            generateTransitionProbabilities(i,
                                            mClusterDistances,
                                            mPreviousMeans,
                                            vech(np.identity(2))*np.ones(iP)[:,None],
                                            dGamma,
                                            dSmooth
                                            )
            
        # Third, generate new firm states
        vCurrentStates = generateNewFirmStates(vCurrentStates, mStateTransitionProbabilities)
        
        # Fourth, update the means
        mPreviousMeans = updateMeans(mPreviousMeans,
                                     iSimulationType,
                                      i,
                                      iCluster2, 
                                      iT,
                                      mCircleCenters,
                                      dCircleRadius,
                                      vTime,
                                      dCircleMidpointDistance,
                                      iTVprob,
                                      dSemiMajorAxis,
                                      iEllipse1Orientation,
                                      iEllipse2Orientation
                                      )      
    return

###############################################################################
####################     Maximum Likelihood functions     #####################
###############################################################################   

def fMaximumLikelihood(cClass, dictEstimationParams):
    """
    PURPOSE:
            Estimate the parameters of the GAS-HMM model using maximum likelihood method. 
    
    INPUTS:
        cClass: cSimulationStudy() object
            Contains storage matrices for data, parameters and states.
            See earlier descriptions for details.
            
        dictEstimationParams: dictionary
            containing params used for estimation.
        
    OUTPUTS:
        mResults: matrix
            contains the stacked estimation results with the True parameters, 
            true states and simulated data.
    """
    iP = dictEstimationParams[0]
    iNrStartingValues = dictEstimationParams[1]
    iPrint = dictEstimationParams[2]
    dInverseDegreesOfFreedom = dictEstimationParams[3]
    dSmooth = dictEstimationParams[4]
    iMeanSpecificSmoothParam = dictEstimationParams[5]
    dWarmStart = dictEstimationParams[6]
    dRegularizeCovariance = dictEstimationParams[7]
    iBurn = dictEstimationParams[8]

    iD = cClass.s_ammData.shape[1]
    iN = cClass.s_ammData.shape[0]
    iT = cClass.s_ammData.shape[2]
    
    mResults = None
        
    # loop over number of starting values
    for iStart in range(iNrStartingValues):
        # initialize      
        vPOriginal, cClass.s_mParameters = fInitializeVp(cClass,
                                                iN,
                                                iP,
                                                iD,
                                                dInverseDegreesOfFreedom,
                                                iMeanSpecificSmoothParam,
                                                dWarmStart)
        

        def fLocalLikelihoodFunction(vP):
            if (np.abs(vP).max(axis=0) >50):
                return 900000000
            
            # transform the parameter vector into the necessary parameters 
            vEwma, dGamma= fGetParameters(cClass,
                                          vP,
                                            iMeanSpecificSmoothParam,
                                            dInverseDegreesOfFreedom)[0:2]

            if (np.abs(np.log(dGamma)) > 5):
                return 900000000
            if (np.isnan(cClass.s_mInitialPredictedProbabilities).any()):
                return 900000000

            
            # run the GAS - HMM - cluster filter
            vLik = fHMMGasFilter(cClass,
                                vEwma,
                                dGamma,
                                cClass.s_mInitialPredictedProbabilities.copy(),
                                dSmooth,
                                dRegularizeCovariance,
                                dInverseDegreesOfFreedom)

            # check whether likelihood can be evaluated
            if (np.isnan(vLik[iBurn:].mean())):
                return 900000000
            else:
                return -vLik[iBurn:].mean()
        
        res= minimize(fLocalLikelihoodFunction,
              x0= vPOriginal,
              method='BFGS',
              options = {'maxiter':20,'gtol': 1e-4})
        
        # set the initial predicted probabilities to the t=1 filtered probabilities
        cClass.s_mInitialPredictedProbabilities = cClass.s_ammFilteredClusterProbabilities[0,:,:].copy()
        
        res= minimize(fLocalLikelihoodFunction,
              x0=vPOriginal,
              method='BFGS',
              options={'maxiter':100,'gtol': 1e-4})
        
    
        print(f'\n BFGS exit status {res.status} and succes:{res.success}.\n{res.message}\n')
        print(res.x)
        print(res)
    

        # switch regimes around if observation 1 does not have the highest tau
        # note that the time paths have not been changed; for this the filter needs to be rerun!       
        if(iP == 2) and (cClass.s_ammFilteredClusterProbabilities[0,0,0] < cClass.s_ammFilteredClusterProbabilities[0,0,1]):
            # change the regime probabilities
            for i in range(iT):
                cClass.s_ammFilteredClusterProbabilities[i,:,:] = (
                (cClass.s_ammFilteredClusterProbabilities[i,:,:].copy()
                 @ np.array([[0,1],
                             [1,0]]))
                )
                cClass.s_ammParameters[i,:,:] = (np.array([[0,1],[1,0]])
                                                 @ cClass.s_ammParameters[i,:,:].copy()
                                                 )
                
        # store the results for this starting value
        if (iStart == 0):
            mResults = [np.hstack((res.fun,
                                    res.status,
                                    res.x,
                                    fRestrictPar(cClass, res.x, iMeanSpecificSmoothParam),
                                    fVecr3dStructure(cClass.s_ammParameters),
                                    fVecr3dStructure(cClass.s_ammFilteredClusterProbabilities)
                                    ))]

        else:
            mResults = [mResults] + [np.hstack((res.fun,
                                                res.status,
                                                res.x,
                                                fRestrictPar(cClass, res.x, iMeanSpecificSmoothParam),
                                                fVecr3dStructure(cClass.s_ammParameters),
                                                fVecr3dStructure(cClass.s_ammFilteredClusterProbabilities)
                                                ))] 
    mResults = np.vstack(mResults)
    vResults = mResults[0,:]
    
    return np.hstack((vResults,
                      fVecr3dStructure(cClass.s_ammTrueParameters),
                      cClass.s_mTrueStates.reshape(-1,1).flatten(),
                      fVecr3dStructure(cClass.s_ammData)
                      ))